; ACL2 Version 1.8

; Copyright (C) 1989-95 Computational Logic, Inc. (CLI).  All rights reserved.

; Use of this software constitutes agreement with the terms of ACL2
; license agreement, found in the file LICENSE.

; This file, acl2-init.lisp, is the inititialization file for ACL2.

; ACL2 is designed to run in any Common Lisp, although we have focused
; especially on AKCL, and we take advantage of a few aspects of AKCL
; to assist our development.  This file need not be distributed with
; ACL2 and is unimportant for the correct operation of ACL2.  This
; file is loaded automatically by ACKL when it starts up.

; This file cannot be compiled because it changes packages in the middle.

#+akcl
(setq si:*notify-gbc* t)

#+lispworks
(lw::extend-current-stack 1000)

; Create the packages we use.

(load "acl2.lisp")

;  Now over to the "ACL2" package for the rest of this file.

(in-package "ACL2")

; If one changes ACL2 in such a way that not even an (initialize-acl2)
; can be executed successfully, then it can be useful to run the
; compiler in a mode that checks thoroughly for calls of nonexistent
; functions or calls with the wrong number of arguments.  To obtain
; such checking in AKCL it is sufficient to change the defparameter
; below to set *emit-fl* to t, in akcl.  A side effect of this
; checking is the creation of a file with the name *.fn when a file
; with the name *.lisp is compiled.  These *.fn files may be deleted
; at any time.

#+akcl
(defparameter *emit-flg* nil)

#+akcl
(cond (*emit-flg* (compiler::emit-fn t)))

; It is a mystery why the following proclamation is necessary, but it
; SEEMS to be necessary in order to permit the interaction of tracing
; with the redefinition of si::break-level.

#+akcl
(proclaim '(special si::arglist))

#+akcl
(let ((v (symbol-function 'si::break-level)))
  (setf (symbol-function 'si::break-level)
        (function
         (lambda (&rest rst)
           (format t "~%Raw Lisp Break.~%")
           (apply v rst)))))

(defun system-call (string arguments)
  #+akcl
  (lisp::system
   (let ((result (concatenate 'string string " ")))
     (dolist
      (x arguments)
      (setq result (concatenate 'string result x)))
     result))
  #+lucid (user::run-program string :arguments arguments)
  #+lispworks
  (system::call-system
   (let ((result (concatenate 'string string " ")))
     (dolist
      (x arguments)
      (setq result (concatenate 'string result x)))
     result)
   "/bin/csh")
  #+allegro (excl::shell 
             (let ((result (concatenate 'string string " ")))
               (dolist
                (x arguments)
                (setq result (concatenate 'string result x)))
               result))
  #-(or akcl lucid lispworks allegro)
  (error "SYSTEM-CALL is not yet defined in this Lisp."))

(defun copy-acl2 (dir)
  (system-call
   "cp"
   (append '(" makefile"
             " acl2.lisp"
             " acl2-check.lisp"
             " acl2-fns.lisp"
             " init.lsp"
             " acl2-init.lisp")
           (append (let ((result (list (format nil " ~a" dir))))
                     (dolist
                      (x *copy-of-acl2-files*)
                      (setq result
                            (cons (format nil " ~a.lisp" x)
                                  result)))
                     result)))))

(defun copy-distribution (output-file
                          source-directory
                          &optional
                          (target-directory "/slocal/src/acl2/acl2-sources"))

; We check that all files and directories exist that are supposed to exist.  We
; cause an error if not, which ultimately will cause the Unix process that
; calls this function to return an error status, thus halting the make of which
; this operation is a part.  Wart:  Since probe-file does not check names with
; wildcards, we skip those.

; Note:  This function does not actually do any copying or directory creation;
; rather, it creates a file that can be executed.

; FIRST, we make sure we are in the expected directory.

  (cond ((not (and (stringp source-directory)
                   (not (equal source-directory ""))))
         (error "The source directory specified for COPY-DISTRIBUTION~%~
                 must be a non-empty string, but~%~s~%is not."
                source-directory)))
  (cond ((eql (char source-directory (1- (length source-directory))) #\/)

; In this code we treat all directories as names without the trailing slash.

         (setq source-directory
               (subseq source-directory 0 (1- (length source-directory))))))
  (cond ((not (equal (format nil "~a/" source-directory)
                     (namestring (truename ""))))
         (error "We expected to be in the directory~%~s~%~
                 but instead are apparently in the directory~%~s .~%~
                 Either issue, in Unix, the command~%~
                 cd ~a~%~
                 or else edit the file (presumably, makefile) from~%~
                 which the function COPY-DISTRIBUTION was called,~%~
                 in order to give it the correct second argument."
                source-directory
                (namestring (truename ""))
                source-directory)))

; Next, check that everything exists that is supposed to.

  (cond ((probe-file target-directory)
         (error "Aborting copying of the distribution.  The target ~%~
                 distribution directory~%~s~%~
                 already exists!  You may wish to execute the following~%~
                 Unix command to remove it and all its contents:~%~
                 rm -r ~a"
                target-directory target-directory)))
  (format t "Checking that distribution files are all present.~%")
  (let (all-dirs missing-files)
    (with-open-file
     (str (concatenate 'string source-directory "/all-files.txt")
          :direction :input)
     (let (filename (dir nil))
       (loop (setq filename (read-line str nil))
             (cond
              ((null filename) (return))
              ((or (equal filename "")
                   (equal (char filename 0) #\#)))
              ((find #\Tab filename)
               (error "Found a line with a Tab in it:  ~s" filename))
              ((find #\Space filename)
               (error "Found a line with a Space in it:  ~s" filename))
              ((find #\* filename)
               (format t "Skipping wildcard file name, ~s.~%" filename))
              ((eql (char filename (1- (length filename))) #\:)

; No need to check for directories here; they'll get checked elsewhere.  But
; it's harmless enough to do so.

               (let* ((new-dir (subseq filename 0 (1- (length filename))))
                      (absolute-dir
                       (format nil "~a/~a" source-directory new-dir)))
                 (cond
                  ((probe-file absolute-dir)
                   (setq dir new-dir))
                  (t
                   (setq missing-files
                         (cons absolute-dir missing-files))
                   (error "Failed to find directory ~a ."
                          absolute-dir)))))
              (t (let ((absolute-filename
                        (if dir
                            (format nil "~a/~a/~a" source-directory dir filename)
                          (format nil "~a/~a" source-directory filename))))
                   (cond
                    ((not (probe-file absolute-filename))
                     (setq missing-files
                           (cons absolute-filename missing-files))
                     (format t "Failed to find file ~a.~%" absolute-filename)))))))))
    (cond
     (missing-files
      (error "~%Missing the following files (and/or directories):~%~s"
             missing-files))
     (t (format t "Distribution files are all present.~%"))))

  (format t "Preparing to copy distribution files from~%~a/~%to~%~a/ .~%"
          source-directory target-directory)
  (let (all-dirs)

; In this pass, we look only for directory names.

    (with-open-file
     (str (concatenate 'string source-directory "/all-files.txt")
          :direction :input)
     (let (filename (dir nil))
       (loop (setq filename (read-line str nil))
             (cond
              ((null filename) (return))
              ((or (equal filename "")
                   (equal (char filename 0) #\#)))
              ((find #\Tab filename)
               (error "Found a line with a Tab in it:  ~s" filename))
              ((find #\Space filename)
               (error "Found a line with a Space in it:  ~s" filename))
              ((eql (char filename (1- (length filename))) #\:)
               (setq all-dirs
                     (cons (subseq filename 0 (1- (length filename)))
                           all-dirs)))))))

; In the final pass we do our writing.

    (with-open-file
     (str (concatenate 'string source-directory "/all-files.txt")
          :direction :input)
     (with-open-file
      (outstr output-file :direction :output)
      (let (filename (dir nil))
        (format outstr "mkdir ~a~%~%" target-directory)
        (loop (setq filename (read-line str nil))
              (cond
               ((null filename) (return))
               ((or (equal filename "")
                    (equal (char filename 0) #\#)))
               ((eql (char filename (1- (length filename))) #\:)
                (setq dir (subseq filename 0 (1- (length filename))))
                (format outstr "~%mkdir ~a/~a~%"
                        target-directory dir))
               ((null dir)
                (cond ((not (member filename all-dirs
                                    :test 'equal))
                       (format outstr "cp -p ~a/~a ~a~%"
                               source-directory
                               filename
                               target-directory))))
               (t
                (cond ((not (member (format nil "~a/~a"
                                            dir filename)
                                    all-dirs
                                    :test 'equal))
                       (format outstr "cp -p ~a/~a/~a ~a/~a~%"
                               source-directory
                               dir
                               filename
                               target-directory
                               dir)))))))))

    (format t "Finished creating a command file for copying distribution files."
            source-directory target-directory)))

(defun make-tags ()
  (system-call "etags"
               (append '(" acl2.lisp"
                         " acl2-check.lisp"
                         " acl2-fns.lisp"
                         " init.lsp"
                         " acl2-init.lisp"
                         " akcl-acl2-trace.lisp"
                         " p.lisp")
                       (let ((result nil))
                         (dolist
                          (x *copy-of-acl2-files*)
                          (setq result
                                (cons (format nil " ~a.lisp" x)
                                      result)))
                         (reverse result)))))

(defun no-object-file-or-out-of-date-object-file (fl)
  (or (null (probe-file (make-pathname :name fl :type *compiled-file-extension*)))
      (> (file-write-date (make-pathname :name fl :type *lisp-extension*))
         (file-write-date (make-pathname :name fl :type *compiled-file-extension*)))))

(defun quick-compile-acl2 (&optional very-fast (small-p t))

; Here is a natural place to put compiler options.

; (proclaim '(optimize (safety 0) (space 0) (speed 3)))

  (maybe-push-small-acl2-image-onto-features small-p)
  (cond
   ((or (not (probe-file (make-pathname :name "acl2-status"
                                        :type "txt")))
        (with-open-file (str (make-pathname :name "acl2-status"
                                            :type "txt")
                             :direction :input)
                        (not (eq (read str nil)
                                 :checked))))
    (check-suitability-for-acl2)))
  (our-with-compilation-unit
   (let ((compile-rest-flg nil)
         (*readtable* *acl2-readtable*)
         #+akcl
         (si:*notify-gbc* nil)
         #+akcl

; AKCL compiler note stuff.  We have so many tail recursive functions
; that the notes about tail recursion optimization are just too much
; to take.

         (compiler:*suppress-compiler-notes* t)
         (files (remove "defpkgs" *copy-of-acl2-files* :test #'equal)))
     (cond
      ((some #'no-object-file-or-out-of-date-object-file files)
       (make-tags)
       #+akcl
       (cond (*emit-flg*
              (dolist (fl files)
                      (let ((file (make-pathname :name fl :type "fn")))
                        (when (probe-file file)
                              (load file))))))
       (dolist
        (fl files)
        (let ((source (make-pathname :name fl :type *lisp-extension*))
              (object (make-pathname :name fl :type *compiled-file-extension*)))
          (cond
           ((equal fl "proof-checker-pkg")
            (load source))
           ((or compile-rest-flg (no-object-file-or-out-of-date-object-file fl))
            (load source)
            (proclaim-file source)
            (when (not very-fast)
                  (setq compile-rest-flg t))
            (compile-file source)
            (load object))
           (t (load object)
              (proclaim-file source)))))
       #+akcl
       (cond (*emit-flg*
              (dolist (fl (cdr files))
                      (let ((file (make-pathname :name fl :type "fn")))
                        (when (probe-file file)
                              (load file))))
              (let ((l (compiler::list-undefined-functions)))
                (cond (l (format t "~%~%Undefined functions: ~a." l))))))
       )
      (t "Nothing to do."))))
  (note-compile-ok))

(defvar *saved-build-date*)
(defvar *saved-mode*)

(defvar *saved-string*
  (concatenate
   'string
   "~% ~a built ~a.~
    ~a"
   #-akcl
   "~% Type (LP) to enter the ACL2 command loop;~
    ~% then see :doc note~a for recent changes.~%"
   #+akcl
   "~% See :doc note~a for recent changes.~%"

   "~% NOTE!!  Proof trees are disabled in ACL2.  To enable them in emacs,~
    ~% look under the ACL2 source directory in interface/emacs/README.doc; ~
    ~% and, to turn on proof trees, execute :START-PROOF-TREE in the ACL2 ~
    ~% command loop.   Look in the ACL2 documentation under PROOF-TREE.~%"))

(defun saved-build-date-string ()
  (with-output-to-string
   (str)
   (setf (get 'tmp-channel *open-output-channel-key*)
         str)
   (print-idate (idate)
                'tmp-channel
                *the-live-state*)
   (remprop 'tmp-channel *open-output-channel-key*)
   str))

#+akcl
(defun save-acl2-in-akcl (sysout-name &optional mode small-p)
  (setq *acl2-allocation-alist*

; If *acl2-allocation-alist* is rebound before allocation is done in
; si::*top-level-hook*, e.g., if it is bound in one's init.lsp or acl2-init.lsp
; file, then such binding will override this one.  The package name shouldn't
; matter for the keys in user's alist, but in the code below we need to keep
; 'hole in the ACL2 package because we refer to it below.

; Where did these numbers come from?  At CLInc we have used the numbers from
; the non-small-p case for some time, and they seem satisfactory.  When we
; moved to a "small" image in Version 1.8, we wanted to have roughly the same
; number of free cells as we've had all along, as a default.  The cons number
; below is obtained by seeing how many pages we had free (pages in use
; multiplied by percent free, as shown by (room)) the last time we built ACL2,
; before modifying ACL2 to support small images, and adding that to the number
; of pages in use in the small image when no extra pages were allocated at
; start-up.  The total was 2917, so that is what we use below.  The relocatable
; size is rather arbitrary, and the hole size has been suggested by Bill
; Schelter.  Finally, the other numbers were unchanged when we used the same
; algorithm described above for cons (except for fixnum, which came out to 99
; -- close enough!).

; Warning:  as of this writing (5/95), there are versions of Linux in which the
; page size is half of that in GCL on a Sparc.  In that case, we should double
; the number of pages in each case in order to have the same amount of free
; objects available.  We do this below; see the next comment.  (We assume that
; there are still the same number of bytes per object; at least, in one
; instance in Linux that appears to be the case for cons, namely, 12 bytes per
; cons.)

        `((hole . 500)
          (relocatable . ,(if small-p 500 1000))
          (cons . ,(if small-p 2917 4500))
          (fixnum . 100)
          (symbol . 400)
          (package)
          (array)
          (string . 200)
          (cfun . 32)
          (sfun . 40)))

; Now adjust if the page size differs from that for GCL/AKCL running on a
; Sparc.  See comment above.

  (let ((multiplier (/ 4096 si::lisp-pagesize)))
    (cond ((not (= multiplier 1))
           (setq *acl2-allocation-alist*
                 (sloop::sloop for (type . n) in *acl2-allocation-alist*
                               collect
                               (cons type
                                     (and n
                                          (round (* multiplier n)))))))))
  (setq si::*top-level-hook*
        #'(lambda ()
            (format t *saved-string*
                    *copy-of-acl2-version*
                    *saved-build-date*
                    (cond (mode
                           (format nil "~% Initialized with ~a." mode))
                          (t ""))
                    (subseq *copy-of-acl2-version*
                            (1+ (search "." *copy-of-acl2-version*))))
            (let ((fl (probe-file (merge-pathnames (user-homedir-pathname)
                                                   "acl2-init.lsp"))))
              (if fl (load fl)))
            (cond
             (*acl2-allocation-alist*
;              (format
;               t
;               "Beginning allocations.  Set acl2::*acl2-allocation-alist* to NIL~%~
;                in ~~/acl2-init.lsp if you must make your running image smaller.~%")
              (sloop::sloop for (type . n) in *acl2-allocation-alist*
                            when n
                            do
;                            (format t "Allocating ~s to ~s.~%" type n)
                            (let ((x (symbol-name type)))
                              (cond
                               ((equal x "HOLE")
                                (si::set-hole-size n))
                               ((equal x "RELOCATABLE")
                                (si::allocate-relocatable-pages n))
                               (t (si::allocate type n t)))))))
            (in-package "ACL2")
            (lp)))
  (load "akcl-acl2-trace.lisp")

; The following is important so that ACL2 functions are efficient in certain
; situations.  For example, (aref1 'foo foo n) should avoid boxing a fixnum n.

  (proclaim-files)

; Return to normal allocation growth.  Keep this in sync with load-acl2, which
; had presumably already set the allocation growth to be particularly slow.

  (sloop::sloop
   for type in
   '(cons fixnum symbol array string cfun sfun

; In akcl, at least some versions of it, we cannot call allocate-growth on the
; following two types.

          #+gcl contiguous
          #+gcl relocatable
          )
   do
   (cond
    ((or (boundp 'si::*gcl-major-version*) ;GCL 2.0 or greater
         (and (boundp 'si::*gcl-version*) ;GCL 1.1
              (= si::*gcl-version* 1)))
     (si::allocate-growth type 0 0 0 0))
    (t (si::allocate-growth type 0 0 0))))

;  (print "Start (si::gbc nil)") ;debugging GC
  (si::set-hole-size 500) ; wfs suggestion
  (si::gbc nil) ; wfs suggestion
;  (print "Start (si::sgc-on t)") ;debugging GC
  (cond ((fboundp 'si::sgc-on) ; si::sgc-on is not bound in Linux version, e.g.
         (si::sgc-on t))) ; wfs suggestion
;  (print "Finish (si::sgc-on t)") ;debugging GC

; Set the hole to be sufficiently large so that ACL2 can do all the allocations
; quickly when it starts up, without any GC, leaving the desired size hole when
; finished.

  (let ((new-hole-size
         (or (cdr (assoc 'hole *acl2-allocation-alist*))
             (si::get-hole-size))))
    (sloop::sloop for (type . n) in *acl2-allocation-alist*
                  with space
                  when (and n
                            (not (equal (symbol-name type) "HOLE"))
                            (< (setq space
                                     #+gcl
                                     (cond ;2.0 or later?
                                      ((boundp 'si::*gcl-major-version*)
                                       (nth 1 (multiple-value-list
                                               (si::allocated type))))
                                      (t
                                       (caddr (si::allocated type))))
                                     #-gcl
                                     (cond
                                      ((equal (symbol-name type)
                                              "RELOCATABLE")
                                       (si::allocated-relocatable-pages))
                                      (t (si::allocated-pages type))))
                               n))
                  do (setq new-hole-size (+ new-hole-size (- n space))))
;    (print "Set hole size") ;debugging
    (si::set-hole-size new-hole-size))

;  (print (true-listp (w *the-live-state*))) ;swap in the world's pages

;  (print "Save the system") ;debugging
  (si::save-system sysout-name))

(defun acl2-default-restart ()
  (format t *saved-string*
          *copy-of-acl2-version*
          *saved-build-date*
          (cond (*saved-mode*
                 (format nil "~% Initialized with ~a." *saved-mode*))
                (t ""))
          (subseq *copy-of-acl2-version*
                  (1+ (search "." *copy-of-acl2-version*))))
  (let ((fl (probe-file (merge-pathnames (user-homedir-pathname)
                                         "acl2-init.lsp"))))
    (if fl (load fl)))
  (in-package "ACL2")
  #+lispworks
  (lw::extend-current-stack 1000)
  nil)

#+lucid
(defun save-acl2-in-lucid (sysout-name &optional mode)
  (setq *saved-mode* mode)
  (user::disksave sysout-name :restart-function 'acl2-default-restart
                  :full-gc t))

#+lispworks
(defun save-acl2-in-lispworks (sysout-name &optional mode)
  (setq *saved-mode* mode)
  (system::save-image sysout-name :restart-function 'acl2-default-restart
                      :gc t))

#+allegro
(defun save-acl2-in-allegro (sysout-name &optional mode)
  (setq *saved-mode* mode)
  (excl:dumplisp :name sysout-name :checkpoint t
                 :restart-function 'acl2-default-restart))

(defun save-acl2 (&optional mode)

; Consider adding something like
; (ccl::save-application "acl2-image" :size (expt 2 24))
; for the Mac.

  (let ((small-p (and (consp mode)
                      (eq (car mode) 'initialize-acl2)
                      (or (null (cdddr mode))
                          (and (cadddr mode)
                               (not (equal (cadddr mode) ''nil)))))))
    (load-acl2 small-p)
    (setq *saved-build-date* (saved-build-date-string))
    (eval mode)
    (cond
     ((or (not (probe-file *acl2-status-file*))
          (with-open-file (str *acl2-status-file*
                               :direction :input)
                          (not (eq (read str nil)
                                   :initialized))))
      (error "Initialization has failed.  Try INITIALIZE-ACL2 again.")))

    #+akcl
    (save-acl2-in-akcl "nsaved_acl2" mode small-p)
    #+lucid
    (save-acl2-in-lucid "nsaved_acl2" mode)
    #+lispworks
    (save-acl2-in-lispworks "nsaved_acl2" mode)
    #+allegro
    (save-acl2-in-allegro "nsaved_acl2" mode)
    #-(or akcl lucid lispworks allegro)
    (error "We do not know how to save ACL2 in this Common Lisp.")))

(defun proclaim-files ()

; IMPORTANT:  This function assumes that the defconst forms in the
; given files have already been evaluated.  One way to achieve this
; state of affairs, of course, is to load the files first.

  (dolist (fl *copy-of-acl2-files*)
          (proclaim-file (format nil "~a.lisp" fl))))

(defun acl2 nil
  (let ((*readtable* *acl2-readtable*))
    (dolist (name (remove "defpkgs" *copy-of-acl2-files* :test #'equal))
            (if (equal name "proof-checker-pkg")
                (load "proof-checker-pkg.lisp")
              (load (make-pathname :name name
                                   :type *compiled-file-extension*))))
    (load "defpkgs.lisp")
    (in-package "ACL2")
    "ACL2"))

(defun user::acl2 ()
  (acl2))

; The following avoids core being dumped in certain circumstances
; resulting from very hard errors.

#+akcl
(si::catch-fatal 1)
