; ACL2 Version 1.8

; Copyright (C) 1989-95 Computational Logic, Inc. (CLI).  All rights reserved.

; Use of this software constitutes agreement with the terms of ACL2
; license agreement, found in the file LICENSE.

; We don't intend this file to be compiled.

;                            TRACE stuff

(defvar *traced-fn*)

; We put over into old-trace the macro for trace that comes with ACKL.
; Thus one can type (old-trace foo) and get the effect that (trace
; foo) would have previously provided.

(cond ((null (macro-function 'old-trace))
       (setf (macro-function 'old-trace) (macro-function 'trace))))

(cond ((null (macro-function 'old-untrace))
       (setf (macro-function 'old-untrace) (macro-function 'untrace))))

(defparameter *trace-evisc-tuple*
  '(nil nil nil nil nil nil))

; Trace prints using the Lisp values of *print-level* and
; *print-length*.

(defun trace-evisc-tuple ()
  (if (and (eq (caaar *trace-evisc-tuple*)
               (w *the-live-state*))
           (eql *print-level* (cadddr *trace-evisc-tuple*))
           (eql *print-length* (car (cddddr *trace-evisc-tuple*))))
      *trace-evisc-tuple*
    (reset-trace-evisc-tuple)))

(defun reset-trace-evisc-tuple ()
  (setq *trace-evisc-tuple*
        (list (world-evisceration-alist *the-live-state* nil)
              (cadr *trace-evisc-tuple*)
              (caddr *trace-evisc-tuple*)
              *print-level*
              *print-length*
              (cadr (cddddr *trace-evisc-tuple*)))))

; Each time one traces a function, the *trace-evisc-tuple* stores the
; current *print-level* and *print-length* with the current world.

(defmacro trace (&rest fns)
  (let (acl2-fns all-fns)
    (sloop::sloop for fn in fns
                  do (if (function-symbolp
                          (cond
                           ((symbolp fn) fn)
                           ((and (consp fn) (symbolp (car fn)))
                            (car fn))
                           (t (error "Not a symbol or a cons of a symbol: ~s" fn)))
                          (w *the-live-state*))
                         (push fn acl2-fns)
                       (push fn all-fns)))
    (cons
     'old-trace
     (progn
       (dolist (x acl2-fns)
         (push (cond ((symbolp x) (*1*-symbol x))
                     (t (cons (*1*-symbol (car x))
                              (cdr x))))
               acl2-fns))
       (dolist (fn acl2-fns)
         (push (progn (cond ((symbolp fn)
                             (setq fn (list fn))))
                      (cond ((and (member :break (cdr fn))
                                  (member :cond (cdr fn)))
                             (interface-er "Use of both :break and :cond is ~
                                            not permitted in TRACE.")))
                      (cons (car fn)
                            (trace-fix-break
                             (car fn)
                             (trace-fix-entry
                              (car fn)
                              (trace-fix-exit (car fn)
                                              (cdr fn))))))
               all-fns))
       all-fns))))

(defmacro untrace (&rest fns)
  (cons
   'old-untrace
   (let ((ans fns))
     (dolist (fn fns)
       (push (*1*-symbol fn) ans))
     ans)))

(defun trace-ppr (x)
  (let ((trace-evisc-tuple (trace-evisc-tuple)))
    (ppr (eviscerate x
                     (cadddr trace-evisc-tuple)
                     (car (cddddr trace-evisc-tuple))
                     (car trace-evisc-tuple)
                     (car (last trace-evisc-tuple)))
         (min (+ 3
                 (* si::*trace-level* 2))
              20)
         'acl2-output-channel::standard-character-output-0
         *the-live-state*
         t)
    '>))

(defparameter *break-package* *acl2-package*)

(defun trace-fix-break (name l)
  (cond ((endp l)
         nil)
        ((eq (car l) :break)
         (list* :cond
                `(let ((*traced-fn* ',name))
                   (let ((*package* *break-package*))
                     (break "at call of ~a.  Args in si::arglist.  ~
                          :r to resume. ~
                          ~%formals:~{ ~a~}.  (e ...) to eval."
                            ',name (trace-formals ',name)))
                   ,(cadr l))
                (cddr l)))
        (t (list* (car l) (cadr l) (trace-fix-break name (cddr l))))))

(defun trace-fix-entry (name l)
  (cond ((endp l)
         (list :entry
               `(trace-ppr (cons ',name (trace-hide-world-and-state
                                         si::arglist)))))
        ((eq (car l) :entry)
         (list* :entry
                `(trace-ppr (cons ',name
                                  (progn (setq si::arglist
                                               (trace-hide-world-and-state
                                                si::arglist))
                                         ,(cadr l))))
                (cddr l)))
        (t (list* (car l) (cadr l) (trace-fix-entry name (cddr l))))))

(defun trace-multiplicity (name)
  (cond ((and (f-boundp-global 'current-acl2-world *the-live-state*)
              (getprop name 'multiplicity nil 'current-acl2-world
                       (f-get-global 'current-acl2-world *the-live-state*))))))

(defun trace-formals (name)
  (cond ((and (f-boundp-global 'current-acl2-world *the-live-state*)
              (getprop name 'formals nil 'current-acl2-world
                       (f-get-global 'current-acl2-world *the-live-state*))))))

(defun trace-values (name)
  (list 'cons
        '(car values)
        (let ((mul (trace-multiplicity name)))
          (cond (mul (list 'acl2::mv-refs (1- mul)))
                (t nil)))))

(defun trace-fix-exit (name l)
  (cond ((endp l)
         (list :exit
               `(trace-ppr (cons ',name
                                 (trace-hide-world-and-state
                                  ,(trace-values name))))))
        ((eq (car l) :exit)
         (list* :exit
                `(trace-ppr (cons ',name
                                  (let
                                      ((values (trace-hide-world-and-state
                                                ,(trace-values name))))
                                    ,(cadr l))))
                (cddr l)))
        (t (list* (car l) (cadr l) (trace-fix-exit name (cddr l))))))

